// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import org.apache.kafka.clients.consumer.internals.PartitionAssignor.Assignment;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.connect.connector.Task;
import org.apache.kafka.connect.errors.ConnectException;
import org.apache.kafka.connect.source.SourceConnector;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import io.confluent.connect.replicator.util.TrialPeriod;
import io.confluent.connect.replicator.util.Version;
import kafka.utils.ZkUtils;

/**
 * The replicator connector is responsible for monitoring the source cluster for topic metadata
 * changes affecting replicated topics and for distributing partition assignments to tasks. Note
 * that this includes matching the whitelisted and blacklisted topics from the configuration
 * against the topics from the source cluster. When new topics are found that need to be replicated,
 * or when partitions are added to topics undergoing replication, this connector will generate
 * a new partition assignment and distribute it to the connector tasks through a Connect rebalance.
 *
 * Note that all of the real work is done in a separate thread (see {@link TopicMonitorThread}).
 */
public class ReplicatorSourceConnector extends SourceConnector {
  private static final Logger log = LoggerFactory.getLogger(ReplicatorSourceConnector.class);
  private static final int MAX_THREAD_JOIN_TIME_MS = 5000;

  private TopicMonitorThread monitorThread;
  private ReplicatorSourceConnectorConfig config;

  public ReplicatorSourceConnector() {
    // no-arg constructor needed for framework
  }

  ReplicatorSourceConnector(ReplicatorSourceConnectorConfig config,
                            TopicMonitorThread monitorThread) {
    this.config = config;
    this.monitorThread = monitorThread;
  }

  @Override
  public String version() {
    return Version.getVersion();
  }

  @Override
  public void start(Map<String, String> props) {
    this.config = new ReplicatorSourceConnectorConfig(props);
    log.info("Starting replicator connector {}", config.getName());

    if (config.isTrial() && !startOrVerifyTrialPeriod(config)) {
      throw new ConnectException("Trial period expired");
    }

    this.monitorThread = new TopicMonitorThread(context, this.config);

    // make this thread a daemon so that it doesn't prevent Connect shutdown if it hangs
    this.monitorThread.setDaemon(true);
    this.monitorThread.start();
  }

  @Override
  public Class<? extends Task> taskClass() {
    return ReplicatorSourceTask.class;
  }

  @Override
  public List<Map<String, String>> taskConfigs(int maxTasks) {
    List<Map<String, String>> taskConfigs = new ArrayList<>();

    Map<String, Assignment> assignments = monitorThread.assignments(maxTasks);
    for (Map.Entry<String, Assignment> assignment : assignments.entrySet()) {
      ReplicatorSourceTaskConfig taskConfig = ReplicatorSourceTaskConfig
              .builder(config)
              .setTaskId(assignment.getKey())
              .setAssignment(assignment.getValue())
              .build();
      taskConfigs.add(taskConfig.originalsStrings());
    }

    return taskConfigs;
  }

  @Override
  public void stop() {
    log.info("Shutting down replicator connector {}", config.getName());

    if (monitorThread != null) {
      monitorThread.shutdown();
      try {
        monitorThread.join(MAX_THREAD_JOIN_TIME_MS);
        if (monitorThread.isAlive())
          log.warn("Failed to shutdown topic monitor thread for connector {}", config.getName());
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
      }
    }
  }

  @Override
  public ConfigDef config() {
    return ReplicatorSourceConnectorConfig.CONFIG_DEF;
  }

  private boolean startOrVerifyTrialPeriod(ReplicatorSourceConnectorConfig config) {
    final ZkUtils zkUtils = config.buildDestZkUtils();
    try {
      return TrialPeriod.startOrVerify(System.currentTimeMillis(), zkUtils);
    } finally {
      zkUtils.close();
    }
  }

}
