// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import org.apache.kafka.common.config.AbstractConfig;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.common.config.ConfigDef.Importance;
import org.apache.kafka.common.config.ConfigDef.Range;
import org.apache.kafka.common.config.ConfigDef.Recommender;
import org.apache.kafka.common.config.ConfigDef.Type;
import org.apache.kafka.common.config.ConfigDef.Width;
import org.apache.kafka.common.config.ConfigException;
import org.apache.kafka.common.record.TimestampType;
import org.apache.kafka.common.security.JaasUtils;
import org.apache.kafka.connect.storage.Converter;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import io.confluent.connect.replicator.util.ByteArrayConverter;
import io.confluent.connect.replicator.util.LicenseValidation;
import kafka.utils.ZkUtils;

public class ReplicatorSourceConnectorConfig extends AbstractConfig {

  public static final String LICENSE_CONFIG = "confluent.license";
  public static final String SRC_KEY_CONVERTER_CONFIG = "src.key.converter";
  public static final String SRC_VALUE_CONVERTER_CONFIG = "src.value.converter";
  public static final String TOPIC_WHITELIST_CONFIG = "topic.whitelist";
  public static final String TOPIC_BLACKLIST_CONFIG = "topic.blacklist";
  public static final String TOPIC_REGEX_CONFIG = "topic.regex";
  public static final String TOPIC_RENAME_FORMAT_CONFIG = "topic.rename.format";
  public static final String TOPIC_POLL_INTERVAL_MS_CONFIG = "topic.poll.interval.ms";
  public static final String SRC_ZOOKEEPER_CONNECT_CONFIG = "src.zookeeper.connect";
  public static final String SRC_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG = "src.zookeeper.session.timeout.ms";
  public static final String SRC_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG = "src.zookeeper.connection.timeout.ms";
  public static final String DST_ZOOKEEPER_CONNECT_CONFIG = "dest.zookeeper.connect";
  public static final String DST_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG = "dest.zookeeper.session.timeout.ms";
  public static final String DST_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG = "dest.zookeeper.connection.timeout.ms";
  public static final String TOPIC_CREATE_BACKOFF_MS_CONFIG = "topic.create.backoff.ms";
  public static final String TOPIC_TIMESTAMP_TYPE_CONFIG = "topic.timestamp.type";
  public static final String TOPIC_AUTO_CREATE_CONFIG = "topic.auto.create";
  public static final String TOPIC_CONFIG_SYNC_CONFIG = "topic.config.sync";
  public static final String TOPIC_CONFIG_SYNC_INTERVAL_MS_CONFIG = "topic.config.sync.interval.ms";
  public static final String TOPIC_PRESERVE_PARTITIONS_CONFIG = "topic.preserve.partitions";

  public static final ConfigDef CONFIG_DEF = baseConfigDef();

  private final String name;

  protected static ConfigDef baseConfigDef() {
    final ConfigDef configDef = new ConfigDef();

    {
      final String group = "Confluent Platform";
      int orderInGroup = 0;
      configDef
          .define(LICENSE_CONFIG,
                  Type.STRING,
                  "",
                  new LicenseValidator(),
                  Importance.HIGH,
                  "Confluent will issue a license key to each subscriber. "
                  + "The license key will be a short snippet of text that you can copy and paste. "
                  + "Without the license key, you can use the Replicator for a 30-day trial period. "
                  + "If you are a subscriber, please contact Confluent Support for more information.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "License");
    }

    {
      final String group = "Source Topics";
      int orderInGroup = 0;
      configDef
          .define(TOPIC_REGEX_CONFIG,
                  Type.STRING,
                  null,
                  Importance.HIGH,
                  "Regex of topics to replicate to the destination cluster.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "Regex")
          .define(TOPIC_WHITELIST_CONFIG,
                  Type.LIST,
                  Collections.emptyList(),
                  Importance.HIGH,
                  "Whitelist of topics to be replicated.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "Whitelist")
          .define(TOPIC_BLACKLIST_CONFIG,
                  Type.LIST,
                  Collections.emptyList(),
                  Importance.HIGH,
                  "Topics to exclude from replication.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "Blacklist")
          .define(TOPIC_POLL_INTERVAL_MS_CONFIG,
                  Type.INT,
                  120 * 1000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "How often to poll the source cluster for new topics matching `topic.whitelist` or `topic.regex`.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Topic Config Sync Interval (ms)");
    }

    {
      final String group = "Source Data Conversion";
      int orderInGroup = 0;
      configDef
          .define(SRC_KEY_CONVERTER_CONFIG,
                  Type.CLASS,
                  ByteArrayConverter.class.getName(),
                  Importance.LOW,
                  "Converter for the key field of messages retrieved from the source cluster.",
                  group,
                  ++orderInGroup,
                  Width.MEDIUM,
                  "Key Converter")
          .define(SRC_VALUE_CONVERTER_CONFIG,
                  Type.CLASS,
                  ByteArrayConverter.class.getName(),
                  Importance.LOW,
                  "Converter for the value field of messages retrieved from the source cluster.",
                  group,
                  ++orderInGroup,
                  Width.MEDIUM,
                  "Value Converter");
    }

    {
      final String group = "Source Zookeeper";
      int orderInGroup = 0;
      configDef
          .define(SRC_ZOOKEEPER_CONNECT_CONFIG,
                  Type.STRING,
                  Importance.HIGH,
                  "Zookeeper connection string for the source cluster.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "Connection String")
          .define(SRC_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG,
                  Type.INT,
                  6000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "Connection timeout in milliseconds for the source Zookeeper cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Connection Timeout (ms)")
          .define(SRC_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG,
                  Type.INT,
                  6000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "Session timeout in milliseconds for the source Zookeeper cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Session Timeout (ms)");
    }

    SourceKafkaConfigs.addDefinitions(configDef);

    {
      final String group = "Destination Topics";
      int orderInGroup = 0;
      configDef
          .define(TOPIC_RENAME_FORMAT_CONFIG,
                  Type.STRING,
                  "${topic}",
                  Importance.HIGH,
                  "A format string for the topic name in the destination cluster, " +
                  "which may contain '${topic}' as a placeholder for the originating topic name. For example, ``dc_${topic}`` " +
                  "for the topic 'orders' will map to the destination topic name 'dc_orders'.\n" +
                  "Be careful of the potential for topic name collisions when configuring replicators from multiple source clusters. " +
                  "We typically recommend that each cluster be given a distinct prefix or suffix (as in the example above).",
                  group,
                  ++orderInGroup,
                  Width.MEDIUM,
                  "Rename Format")
          .define(TOPIC_AUTO_CREATE_CONFIG,
                  Type.BOOLEAN,
                  true,
                  Importance.LOW,
                  "Whether to automatically create topics in the destination cluster if required.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Auto-create")
          .define(TOPIC_PRESERVE_PARTITIONS_CONFIG,
                  Type.BOOLEAN,
                  true,
                  Importance.LOW,
                  "Whether to automatically increase the number of partitions in the destination cluster to match the " +
                  "source cluster and ensure that messages replicated from the source cluster use the same partition " +
                  "in the destination cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Preserve Partitions")
          .define(TOPIC_CREATE_BACKOFF_MS_CONFIG,
                  Type.INT,
                  120000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "Time to wait before retrying auto topic creation or expansion.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Creation Backoff (ms)",
                  new BooleanParentRecommender(TOPIC_AUTO_CREATE_CONFIG))
          .define(TOPIC_CONFIG_SYNC_CONFIG,
                  Type.BOOLEAN,
                  true,
                  Importance.LOW,
                  "Whether to periodically sync topic configuration to the destination cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Enable Topic Config Sync")
          .define(TOPIC_CONFIG_SYNC_INTERVAL_MS_CONFIG,
                  Type.INT,
                  120 * 1000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "How often to check for configuration changes when ``topic.config.sync`` is enabled.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Topic Config Sync Interval (ms)",
                  new BooleanParentRecommender(TOPIC_CONFIG_SYNC_CONFIG))
          .define(TOPIC_TIMESTAMP_TYPE_CONFIG,
                  Type.STRING,
                  TimestampType.CREATE_TIME.toString(),
                  ConfigDef.ValidString.in(TimestampType.CREATE_TIME.toString(), TimestampType.LOG_APPEND_TIME.toString()),
                  Importance.LOW,
                  "The timestamp type for the topics in the destination cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Timestamp Type",
                  new BooleanParentRecommender(TOPIC_AUTO_CREATE_CONFIG, TOPIC_CONFIG_SYNC_CONFIG));
    }

    {
      final String group = "Destination Zookeeper";
      int orderInGroup = 0;
      configDef
          .define(DST_ZOOKEEPER_CONNECT_CONFIG,
                  Type.STRING,
                  Importance.HIGH,
                  "Zookeeper connection string for the destination cluster.",
                  group,
                  ++orderInGroup,
                  Width.LONG,
                  "Connection String")
          .define(DST_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG,
                  Type.INT,
                  6000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "Connection timeout in milliseconds for the destination Zookeeper cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Connection Timeout (ms)")
          .define(DST_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG,
                  Type.INT,
                  6000,
                  Range.atLeast(0),
                  Importance.LOW,
                  "Session timeout in milliseconds for the destination Zookeeper cluster.",
                  group,
                  ++orderInGroup,
                  Width.SHORT,
                  "Session Timeout (ms)");
    }

    return configDef;
  }

  private static class BooleanParentRecommender implements Recommender {
    private final String[] parentConfigNames;

    BooleanParentRecommender(String... parentConfigNames) {
      this.parentConfigNames = parentConfigNames;
    }

    @Override
    public List<Object> validValues(String name, Map<String, Object> connectorConfigs) {
      return Collections.emptyList();
    }

    @Override
    public boolean visible(String name, Map<String, Object> connectorConfigs) {
      for (String parentConfigName : parentConfigNames) {
        if ((Boolean) connectorConfigs.get(parentConfigName)) {
          return true;
        }
      }
      return false;
    }
  }

  private static class LicenseValidator implements ConfigDef.Validator {
    @Override
    public void ensureValid(String configKey, Object o) {
      final String licenseStr = (String) o;
      if (!licenseStr.isEmpty()) {
        try {
          LicenseValidation.verify(licenseStr);
        } catch (LicenseValidation.Failure e) {
          throw new ConfigException(configKey, e);
        }
      }
    }

    @Override
    public String toString() {
      return "Confluent Platform license";
    }
  }

  public ReplicatorSourceConnectorConfig(Map<String, String> parsedConfig) {
    super(CONFIG_DEF, parsedConfig);
    this.name = parseName(parsedConfig);
  }

  protected ReplicatorSourceConnectorConfig(ConfigDef subclassConfigDef, Map<String, String> props) {
    super(subclassConfigDef, props);
    this.name = parseName(props);
  }

  public Converter getSourceKeyConverter() {
    return getConfiguredInstance(SRC_KEY_CONVERTER_CONFIG, Converter.class);
  }

  public Converter getSourceValueConverter() {
    return getConfiguredInstance(SRC_VALUE_CONVERTER_CONFIG, Converter.class);
  }

  public String getName() {
    return name;
  }

  public ZkUtils buildDestZkUtils() {
    String zookeeperConnect = getString(ReplicatorSourceTaskConfig.DST_ZOOKEEPER_CONNECT_CONFIG);
    int sessionTimeoutMs = getInt(ReplicatorSourceTaskConfig.DST_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG);
    int connectionTimeoutMs = getInt(ReplicatorSourceTaskConfig.DST_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG);
    return ZkUtils.apply(zookeeperConnect, sessionTimeoutMs, connectionTimeoutMs, JaasUtils.isZkSecurityEnabled());
  }

  public ZkUtils buildSrcZkUtils() {
    String zookeeperConnect = getString(ReplicatorSourceTaskConfig.SRC_ZOOKEEPER_CONNECT_CONFIG);
    int sessionTimeoutMs = getInt(ReplicatorSourceTaskConfig.SRC_ZOOKEEPER_SESSION_TIMEOUT_MS_CONFIG);
    int connectionTimeoutMs = getInt(ReplicatorSourceTaskConfig.SRC_ZOOKEEPER_CONNECTION_TIMEOUT_MS_CONFIG);
    return ZkUtils.apply(zookeeperConnect, sessionTimeoutMs, connectionTimeoutMs, JaasUtils.isZkSecurityEnabled());
  }

  public Pattern getTopicPattern() {
    String regex = getString(TOPIC_REGEX_CONFIG);
    return regex == null ? null : Pattern.compile(regex);
  }

  public Set<String> getTopics() {
    return new HashSet<>(getList(TOPIC_WHITELIST_CONFIG));
  }

  public Set<String> getBlacklistTopics() {
    return new HashSet<>(getList(TOPIC_BLACKLIST_CONFIG));
  }

  public String getTopicRenameFormat() {
    return getString(TOPIC_RENAME_FORMAT_CONFIG);
  }

  public int getTopicPollIntervalMs() {
    return getInt(TOPIC_POLL_INTERVAL_MS_CONFIG);
  }

  public int getTopicConfigSyncIntervalMs() {
    return getInt(TOPIC_CONFIG_SYNC_INTERVAL_MS_CONFIG);
  }

  public Map<String, ?> getSourceConsumerConfigs() {
    final Map<String, Object> configs = originalsWithPrefix(SourceKafkaConfigs.PREFIX);
    configs.putAll(originalsWithPrefix(SourceKafkaConfigs.CONSUMER_PREFIX));
    return configs;
  }

  public int getTopicCreateBackoffMs() {
    return getInt(TOPIC_CREATE_BACKOFF_MS_CONFIG);
  }

  public boolean getTopicAutoCreate() {
    return getBoolean(TOPIC_AUTO_CREATE_CONFIG);
  }

  public boolean getTopicPreservePartitions() {
    return getBoolean(TOPIC_PRESERVE_PARTITIONS_CONFIG);
  }

  public boolean getTopicConfigSync() {
    return getBoolean(TOPIC_CONFIG_SYNC_CONFIG);
  }

  public String getTopicTimestampType() {
    return getString(TOPIC_TIMESTAMP_TYPE_CONFIG);
  }

  public boolean isTrial() {
    return getString(LICENSE_CONFIG).isEmpty();
  }

  protected static String parseName(Map<String, String> props) {
    String nameProp = props.get("name");
    return nameProp != null ? nameProp : "replicator-source";
  }

  public static void main(String[] args) {
    System.out.println(CONFIG_DEF.toRst());
  }

}
