// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import io.confluent.common.utils.SystemTime;
import io.confluent.common.utils.Time;
import io.confluent.connect.replicator.util.ReplicatorAdminClient;
import io.confluent.connect.replicator.util.ReplicatorAdminClient.TopicMetadataListener;
import io.confluent.connect.replicator.util.TopicMetadata;
import io.confluent.connect.replicator.util.Utils;
import io.confluent.connect.replicator.util.Version;
import kafka.log.LogConfig;
import org.apache.kafka.clients.ClientUtils;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetOutOfRangeException;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.ConfigException;
import org.apache.kafka.common.errors.WakeupException;
import org.apache.kafka.common.record.TimestampType;
import org.apache.kafka.common.serialization.ByteArrayDeserializer;
import org.apache.kafka.connect.data.SchemaAndValue;
import org.apache.kafka.connect.errors.ConnectException;
import org.apache.kafka.connect.source.SourceRecord;
import org.apache.kafka.connect.source.SourceTask;
import org.apache.kafka.connect.source.SourceTaskContext;
import org.apache.kafka.connect.storage.Converter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

import static io.confluent.connect.replicator.util.Utils.toConnectOffset;
import static io.confluent.connect.replicator.util.Utils.toConnectPartition;

/**
 * The task is responsible for consuming from a set of replicated partitions and converting
 * to source records as needed by Connect. On initialization, the task will receive a set
 * of partitions to replicate from the source cluster.
 *
 * The task is responsible for verifying that the topics exist in the destination cluster
 * prior to sending any data from those partitions. To prevent topic creation herding effects,
 * we give the task assigned partition 0 of a particular topic the job of creating or expanding
 * the corresponding topic (after applying topic.rename.format) in the destination cluster.
 *
 * Until a partition has been created, the task will pause the corresponding partition
 * in the source cluster. As soon as the task gets notification that the partition exists
 * (using {@link ReplicatorAdminClient}, the partition will be resumed and replication will
 * begin.
 */
public class ReplicatorSourceTask extends SourceTask {
  private static final Logger log = LoggerFactory.getLogger(ReplicatorSourceTask.class);

  private final Time time;
  private String taskId;
  private ReplicatorSourceTaskConfig config;
  private Converter sourceKeyConverter;
  private Converter sourceValueConverter;
  private ReplicatorAdminClient sourceAdminClient;
  private ReplicatorAdminClient destAdminClient;

  // used to track topics that need creation/expansion
  private final Set<String> sourceTopicsNeedingExpansion = new HashSet<>();
  private Long retryTopicExpansionDeadline;

  // used to track topics which need periodic config checks. This includes all topics for
  // which this task was assigned partition 0.
  private final Set<String> managedSourceTopics = new HashSet<>();
  private Long topicConfigCheckDeadline;

  // enables periodic checks for resuming paused partitions that supplement explicit wakeups invoked by metadata refresh
  // callbacks
  private volatile Long tryResumeDeadline;
  //package-private for testing
  static final long RETRY_RESUME_FREQUENCY_MS = 5000;

  // volatile since it is accessed without the lock (in order to call wakeup)
  private volatile Consumer<byte[], byte[]> consumer;

  public ReplicatorSourceTask() {
    this.time = new SystemTime();
  }

  // visible for testing
  ReplicatorSourceTask(ReplicatorSourceTaskConfig config,
                       SourceTaskContext context,
                       String taskId,
                       Time time,
                       Consumer<byte[], byte[]> consumer,
                       ReplicatorAdminClient sourceAdminClient,
                       ReplicatorAdminClient destAdminClient,
                       Converter sourceKeyConverter,
                       Converter sourceValueConverter) {
    this.config = config;
    this.context = context;
    this.taskId = taskId;
    this.time = time;
    this.consumer = consumer;
    this.sourceAdminClient = sourceAdminClient;
    this.destAdminClient = destAdminClient;
    this.sourceKeyConverter = sourceKeyConverter;
    this.sourceValueConverter = sourceValueConverter;
    this.retryTopicExpansionDeadline = null;
    this.tryResumeDeadline = null;
  }

  @Override
  public String version() {
    return Version.getVersion();
  }

  @Override
  public synchronized void start(Map<String, String> props) {
    try {
      this.config = new ReplicatorSourceTaskConfig(props);
    } catch (ConfigException e) {
      throw new ConnectException("Failed to start Kafka replicator task due to configuration error", e);
    }

    this.taskId = config.getTaskId();
    this.sourceKeyConverter = config.getSourceKeyConverter();
    this.sourceValueConverter = config.getSourceValueConverter();
    this.destAdminClient = new ReplicatorAdminClient(config.buildDestZkUtils(), time, 30000);
    this.sourceAdminClient = new ReplicatorAdminClient(config.buildSrcZkUtils(), time, 30000);
    this.consumer = buildSourceConsumer(config);

    Collection<TopicPartition> sourceAssignment = config.getPartitionAssignment().partitions();
    doStart(sourceAssignment);

    log.info("Started kafka replicator task {} replicating partitions {}", taskId, sourceAssignment);
  }

  synchronized void doStart(Collection<TopicPartition> sourceAssignment) {
    this.consumer.assign(sourceAssignment);
    watchTopicsForAssignedPartitions(sourceAssignment);
    initializeAssignedPartitions(sourceAssignment);
  }

  @Override
  public synchronized List<SourceRecord> poll() throws InterruptedException {
    try {
      // Retry topic creation/expansion to unblock any paused partitions.
      retryTopicExpansionIfNeeded();

      // Periodically check for configuration changes in the source cluster.
      updateTopicConfigsIfNeeded();

      // Attempt to resume any partitions for recently created topics.
      maybeResumePartitions();

      long nextDeadline = Utils.nextDeadline(retryTopicExpansionDeadline, topicConfigCheckDeadline, tryResumeDeadline);
      ConsumerRecords<byte[], byte[]> records = consumer.poll(Math.max(0, nextDeadline - time.milliseconds()));
      if (records.isEmpty())
        return Collections.emptyList();

      // Note that we will only receive records for topics which have already been
      // verified to exist in the destination cluster. Any partitions which do not
      // exist will be paused until topic creation has completed.

      boolean preservePartitions = config.getTopicPreservePartitions();
      List<SourceRecord> sourceRecords = new ArrayList<>(records.count());
      for (TopicPartition topicPartition : records.partitions()) {
        String sourceTopic = topicPartition.topic();
        String destTopic = toDestTopic(sourceTopic);

        int partition = topicPartition.partition();
        Map<String, ?> connectSourcePartition = toConnectPartition(sourceTopic, partition);

        for (ConsumerRecord<byte[], byte[]> record : records.records(topicPartition)) {
          Map<String, Object> connectOffset = toConnectOffset(record.offset());
          SchemaAndValue key = sourceKeyConverter.toConnectData(sourceTopic, record.key());
          SchemaAndValue value = sourceValueConverter.toConnectData(sourceTopic, record.value());
          Long timestamp = record.timestampType() == TimestampType.NO_TIMESTAMP_TYPE ? null : record.timestamp();
          Integer destPartition = preservePartitions ? partition : null;

          sourceRecords.add(new SourceRecord(
                  connectSourcePartition,
                  connectOffset,
                  destTopic,
                  destPartition,
                  key.schema(),
                  key.value(),
                  value.schema(),
                  value.value(),
                  timestamp));
        }
      }

      return sourceRecords;
    } catch (OffsetOutOfRangeException e) {
      // The offset can fall out of range for several reasons including log segment
      // deletion (e.g. if the connector is offline for a long time), or even if a
      // topic is inadvertently deleted in the source cluster. In practice, there's not
      // a lot a user can do when this happens since the next data is already gone, so
      // we simply seek to the beginning of the partition and resume.

      Map<TopicPartition, Long> outOfRangePartitions = e.offsetOutOfRangePartitions();
      log.warn("Consumer from source cluster detected out of range partitions: {}", outOfRangePartitions);
      consumer.seekToBeginning(outOfRangePartitions.keySet());
      return Collections.emptyList();
    } catch (WakeupException e) {
      log.debug("Kafka replicator task {} woken up", taskId);
      return Collections.emptyList();
    }
  }

  private void maybeResumePartitions() {
    if (tryResumeDeadline == null || tryResumeDeadline > time.milliseconds()) {
      return;
    }

    // Check if there are partitions that can be resumed (which will happen when the
    // topic is created or expanded by whichever task owns partition 0 of that topic).
    for (TopicPartition sourcePartition : consumer.paused()) {
      TopicPartition destPartition = toDestPartition(sourcePartition);
      if (destAdminClient.partitionExists(destPartition)) {
        log.debug("Resuming paused partition {} since partition {} now exists in the destination cluster",
                sourcePartition, destPartition);
        consumer.resume(Collections.singleton(sourcePartition));
      }
    }

    if (!consumer.paused().isEmpty()) {
      tryResumeDeadline = time.milliseconds() + RETRY_RESUME_FREQUENCY_MS;
    } else {
      tryResumeDeadline = null;
    }
  }

  private void watchTopicsForAssignedPartitions(Collection<TopicPartition> sourceAssignment) {
    // Collect all the topics in the destination cluster which we need metadata for.
    Set<String> interestedDestTopics = new HashSet<>();
    for (TopicPartition sourcePartition : sourceAssignment)
      interestedDestTopics.add(toDestTopic(sourcePartition.topic()));

    // Get notifications whenever metadata changes for any of these topics
    destAdminClient.setInterestedTopics(interestedDestTopics, new TopicMetadataListener() {
      @Override
      public void onTopicMetadataRefresh() {
        // Wakeup the consumer (which may be in a long poll) to check whether there are any
        // partitions that are able to be resumed.
        tryResumeDeadline = time.milliseconds();
        consumer.wakeup();
      }
    });
  }

  private void initializeAssignedPartitions(Collection<TopicPartition> sourceAssignment) {
    for (TopicPartition sourcePartition : sourceAssignment) {
      TopicPartition destPartition = toDestPartition(sourcePartition);

      if (sourcePartition.partition() == 0) {
        // Whichever task is assigned partition 0 is responsible for verifying that the topic exists, has enough partitions, and topic configuration is in sync.

        String sourceTopic = sourcePartition.topic();

        if (isDestTopicExpansionNeeded(sourceTopic))
          sourceTopicsNeedingExpansion.add(sourceTopic);

        if (config.getTopicConfigSync())
          managedSourceTopics.add(sourceTopic);
      }

      if (config.getTopicPreservePartitions() && !destAdminClient.partitionExists(destPartition)) {
        consumer.seekToBeginning(Collections.singleton(sourcePartition));
        // We must pause the partition until the partition has been created.
        log.info("Seeking to the beginning and pausing source partition {} since destination partition {} does not " +
            "exist yet", sourcePartition, destPartition);
        consumer.pause(Collections.singleton(sourcePartition));
        tryResumeDeadline = time.milliseconds() + RETRY_RESUME_FREQUENCY_MS;
      } else {
        seekSourcePartition(sourcePartition);
      }
    }
  }

  private void seekSourcePartition(TopicPartition sourcePartition) {
    Map<String, ?> connectSourcePartition = toConnectPartition(sourcePartition);
    Map<String, Object> connectSourceOffset = context.offsetStorageReader().offset(connectSourcePartition);
    if (connectSourceOffset == null) {
      log.debug("Seeking to the beginning of source partition {} since there is no committed offset", sourcePartition);
      consumer.seekToBeginning(Collections.singleton(sourcePartition));
    } else {
      long offset = (long) connectSourceOffset.get("offset");
      log.debug("Seeking to committed offset {} for source partition {}", offset, sourcePartition);

      // Note that the consumer expects the position to point to the offset of the next record to be consumed
      consumer.seek(sourcePartition, offset + 1);
    }
  }

  private TopicPartition toDestPartition(TopicPartition sourcePartition) {
    return new TopicPartition(toDestTopic(sourcePartition.topic()), sourcePartition.partition());
  }

  private String toDestTopic(String sourceTopic) {
    return Utils.renameTopic(config.getTopicRenameFormat(), sourceTopic);
  }

  private void retryTopicExpansionIfNeeded() {
    if (sourceTopicsNeedingExpansion.isEmpty()) {
      retryTopicExpansionDeadline = null;
      return;
    }

    if (retryTopicExpansionDeadline == null || retryTopicExpansionDeadline < time.milliseconds()) {
      for (String sourceTopic : sourceTopicsNeedingExpansion)
        maybeCreateOrExpandDestTopic(sourceTopic);
      retryTopicExpansionDeadline = time.milliseconds() + config.getTopicCreateBackoffMs();
    }
  }

  private void overrideTimestampType(Properties sourceTopicConfig) {
    // The timestamp type in source topic config is overridden to match the desired timestamp type in the destination.
    sourceTopicConfig.setProperty(LogConfig.MessageTimestampTypeProp(), config.getTopicTimestampType());
  }

  private void updateTopicConfigsIfNeeded() {
    if (managedSourceTopics.isEmpty()) {
      topicConfigCheckDeadline = null;
      return;
    }

    if (topicConfigCheckDeadline != null && time.milliseconds() < topicConfigCheckDeadline)
      return;

    log.debug("Verifying topic configuration for topics {} for replicator task {}", managedSourceTopics, taskId);

    for (String sourceTopic : managedSourceTopics) {
      String destTopic = toDestTopic(sourceTopic);
      if (!destAdminClient.topicExists(destTopic))
        continue;

      try {
        Properties sourceTopicConfig = sourceAdminClient.topicConfig(sourceTopic);
        overrideTimestampType(sourceTopicConfig);
        Properties destTopicConfig = destAdminClient.topicConfig(destTopic);

        if (!sourceTopicConfig.equals(destTopicConfig)) {
          log.info("Updating configuration of topic {} with properties {}", destTopic, sourceTopicConfig);
          destAdminClient.changeTopicConfig(destTopic, sourceTopicConfig);
        }
      } catch (RuntimeException e) {
        log.warn("Failed topic configuration check for topic {} on task {}. Will retry later.", destTopic, taskId, e);
      }
    }

    topicConfigCheckDeadline = time.milliseconds() + config.getTopicConfigSyncIntervalMs();
  }

  private boolean isDestTopicExpansionNeeded(String sourceTopic) {
    boolean autoCreateTopics = config.getTopicAutoCreate();
    boolean preservePartitions = config.getTopicPreservePartitions();

    if (!autoCreateTopics && !preservePartitions)
      return false;

    List<PartitionInfo> sourcePartitionMetadata = consumer.partitionsFor(sourceTopic);
    int numSourcePartitions = sourcePartitionMetadata.size();
    TopicMetadata destTopicMetadata = destAdminClient.topicMetadata(toDestTopic(sourceTopic));

    if (destTopicMetadata == null)
      return autoCreateTopics;

    return preservePartitions && numSourcePartitions > destTopicMetadata.numPartitions();
  }

  private void maybeCreateOrExpandDestTopic(String sourceTopic) {
    String destTopic = toDestTopic(sourceTopic);
    try {
      List<PartitionInfo> sourcePartitionMetadata = consumer.partitionsFor(sourceTopic);
      int numSourcePartitions = sourcePartitionMetadata.size();

      TopicMetadata destTopicMetadata = destAdminClient.topicMetadata(destTopic);
      if (config.getTopicAutoCreate() && destTopicMetadata == null) {
        // Use the topic configuration from the source cluster when creating the topic.
        // We have to assume that the default broker configuration is the same.
        Properties sourceTopicConfig = sourceAdminClient.topicConfig(sourceTopic);
        overrideTimestampType(sourceTopicConfig);
        // We use the replication factor of the source cluster as advice when creating the
        // topic, but we are limited by the number of brokers actually available.
        int sourceReplicationFactor = sourcePartitionMetadata.get(0).replicas().length;
        int aliveBrokers = destAdminClient.aliveBrokers();

        if (sourceReplicationFactor > aliveBrokers) {
          log.warn("Unable to create topic {} in the destination cluster because the source replication factor {}" +
                  "is greater than the number of active brokers {}", destTopic, sourceReplicationFactor, aliveBrokers);
          return;
        }

        log.info("Creating topic {} in the destination cluster with {} partitions and replication factor {}",
                destTopic, numSourcePartitions, sourceReplicationFactor);
        destAdminClient.createTopic(
                destTopic,
                numSourcePartitions,
                sourceReplicationFactor,
                sourceTopicConfig);
      } else if (config.getTopicPreservePartitions() && (destTopicMetadata != null && numSourcePartitions > destTopicMetadata.numPartitions())) {
        // We need to increase the number of partitions in the destination cluster
        log.info("Increasing number of partitions of topic {} from {} to {} in the destination cluster",
                destTopic, destTopicMetadata.numPartitions(), numSourcePartitions);
        destAdminClient.addPartitions(destTopic, numSourcePartitions);
      }

      // The topic was successfully created/expanded, so remove tracking
      sourceTopicsNeedingExpansion.remove(sourceTopic);
    } catch (RuntimeException e) {
      // If there are any problems in topic creation, we log a warning and retry. Some errors
      // could be considered fatal (e.g. authorization errors), but we have chosen not to fail
      // the task so that replication of other topics can continue. Once Connect adds support
      // for lag tracking, we should have a better way to report problems like this.
      log.warn("Failed to create or expand topic {} in the destination cluster. Will try again later.", destTopic, e);
    }
  }

  @Override
  public void stop() {
    log.info("Closing kafka replicator task {}", taskId);

    if (consumer != null) {
      consumer.wakeup();
      synchronized (this) {
        AtomicReference<Throwable> firstException = new AtomicReference<>();
        ClientUtils.closeQuietly(sourceAdminClient, "source admin client", firstException);
        ClientUtils.closeQuietly(destAdminClient, "destination admin client", firstException);
        ClientUtils.closeQuietly(consumer, "consumer", firstException);
      }
    }
  }

  private KafkaConsumer<byte[], byte[]> buildSourceConsumer(ReplicatorSourceTaskConfig config) {
    Map<String, Object> consumerConfig = new HashMap<>();
    consumerConfig.putAll(config.getSourceConsumerConfigs());
    if (!consumerConfig.containsKey(ConsumerConfig.CLIENT_ID_CONFIG)) {
      consumerConfig.put(ConsumerConfig.CLIENT_ID_CONFIG, taskId);
    }
    consumerConfig.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
    consumerConfig.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "none");
    return new KafkaConsumer<>(consumerConfig, new ByteArrayDeserializer(), new ByteArrayDeserializer());
  }

}
