// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.config.SslConfigs;
import org.apache.kafka.common.protocol.SecurityProtocol;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

public class SourceKafkaConfigs {

  public static final String PREFIX = "src.kafka.";
  public static final String CONSUMER_PREFIX = "src.consumer.";

  private static final Map<String, ConfigDef.ConfigKey> ALL_CONSUMER_CONFIGS;

  static {
    try {
      final Field configField = ConsumerConfig.class.getDeclaredField("CONFIG");
      configField.setAccessible(true);
      final ConfigDef consumerConfigDef = (ConfigDef) configField.get(ConsumerConfig.class);
      ALL_CONSUMER_CONFIGS = new TreeMap<>(consumerConfigDef.configKeys());
    } catch (NoSuchFieldException | IllegalAccessException e) {
      throw new RuntimeException(e);
    }
  }

  public static void addDefinitions(ConfigDef configDef) {
    defineShared(configDef);
    defineSharedSecurity(configDef);
    defineConsumer(configDef);
  }

  private static void defineShared(ConfigDef configDef) {
    final String group = "Source Kafka";
    int orderInGroup = 0;

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.BOOTSTRAP_SERVERS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.LONG);

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.CLIENT_ID_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.MEDIUM);

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.REQUEST_TIMEOUT_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.RETRY_BACKOFF_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.CONNECTIONS_MAX_IDLE_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.RECONNECT_BACKOFF_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.METRIC_REPORTER_CLASSES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.LONG);
    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.METRICS_NUM_SAMPLES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.METRICS_SAMPLE_WINDOW_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.SEND_BUFFER_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.RECEIVE_BUFFER_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
  }

  private static void defineSharedSecurity(ConfigDef configDef) {
    final String group = "Source Kafka: Security";
    int orderInGroup = 0;

    define(configDef, PREFIX, baseConfigKey(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT,
           ConfigDef.ValidString.in(SecurityProtocolRecommender.VALID_VALUES.toArray(new String[SecurityProtocolRecommender.VALID_VALUES.size()])),
           new SecurityProtocolRecommender());

    final SecurityProtocolSpecificRecommender saslRecommender = new SecurityProtocolSpecificRecommender(
        PREFIX + CommonClientConfigs.SECURITY_PROTOCOL_CONFIG,
        SecurityProtocol.SASL_PLAINTEXT, SecurityProtocol.SASL_SSL
    );
    for (ConfigDef.ConfigKey saslConfigKey : saslConfigs().values()) {
      define(configDef, PREFIX, saslConfigKey, group, ++orderInGroup, ConfigDef.Width.NONE, null, saslRecommender);
    }

    final SecurityProtocolSpecificRecommender sslRecommender = new SecurityProtocolSpecificRecommender(
        PREFIX + CommonClientConfigs.SECURITY_PROTOCOL_CONFIG,
        SecurityProtocol.SASL_SSL, SecurityProtocol.SSL
    );
    for (ConfigDef.ConfigKey sslConfigKey : sslConfigs().values()) {
      define(configDef, PREFIX, sslConfigKey, group, ++orderInGroup, ConfigDef.Width.NONE, null, sslRecommender);
    }
  }

  private static void defineConsumer(ConfigDef configDef) {
    final String group = "Source Kafka: Consumer";
    int orderInGroup = 0;

    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.INTERCEPTOR_CLASSES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.LONG);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.FETCH_MAX_WAIT_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.FETCH_MIN_BYTES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.FETCH_MAX_BYTES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.MAX_PARTITION_FETCH_BYTES_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.MAX_POLL_INTERVAL_MS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.MAX_POLL_RECORDS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
    define(configDef, CONSUMER_PREFIX, baseConfigKey(ConsumerConfig.CHECK_CRCS_CONFIG),
           group, ++orderInGroup, ConfigDef.Width.SHORT);
  }

  private static ConfigDef.ConfigKey baseConfigKey(String key) {
    return ALL_CONSUMER_CONFIGS.get(key);
  }

  private static void define(
      ConfigDef configDef,
      String prefix,
      ConfigDef.ConfigKey baseConfigKey,
      String group,
      int orderInGroup,
      ConfigDef.Width width
  ) {
    define(configDef, prefix, baseConfigKey, group, orderInGroup, width, null, null);
  }

  private static void define(
      ConfigDef configDef,
      String prefix,
      ConfigDef.ConfigKey baseConfigKey,
      String group,
      int orderInGroup,
      ConfigDef.Width width,
      ConfigDef.Validator validator,
      ConfigDef.Recommender recommender
  ) {
    configDef.define(
        prefix + baseConfigKey.name,
        baseConfigKey.type,
        baseConfigKey.defaultValue,
        validator != null ? validator : baseConfigKey.validator,
        baseConfigKey.importance,
        htmlToRst(baseConfigKey.documentation),
        group,
        orderInGroup,
        width,
        ".." + baseConfigKey.name,
        recommender
    );
  }

  private static Map<String, ConfigDef.ConfigKey> sslConfigs() {
    final ConfigDef cfg = new ConfigDef();
    SslConfigs.addClientSslSupport(cfg);
    return cfg.configKeys();
  }

  private static Map<String, ConfigDef.ConfigKey> saslConfigs() {
    final ConfigDef cfg = new ConfigDef();
    SaslConfigs.addClientSaslSupport(cfg);
    return cfg.configKeys();
  }

  private static String htmlToRst(String html) {
    return html.trim().replace("<code>", "``").replace("</code>", "``").replace("&mdash;", " — ");
  }

  private static class SecurityProtocolSpecificRecommender implements ConfigDef.Recommender {
    private final String configKey;
    private final Set<String> protocolNames;

    SecurityProtocolSpecificRecommender(String configKey, SecurityProtocol... protocols) {
      this.configKey = configKey;
      this.protocolNames = new HashSet<>(protocols.length);
      for (SecurityProtocol protocol : protocols) {
        protocolNames.add(protocol.name());
      }
    }

    @Override
    public List<Object> validValues(String s, Map<String, Object> map) {
      return Collections.emptyList();
    }

    @Override
    public boolean visible(String s, Map<String, Object> map) {
      return protocolNames.contains(map.get(configKey));
    }
  }

  private static class SecurityProtocolRecommender implements ConfigDef.Recommender {
    static final List<String> VALID_VALUES = new ArrayList<>();

    static {
      for (SecurityProtocol protocol : SecurityProtocol.nonTestingValues()) {
        VALID_VALUES.add(protocol.name());
      }
    }

    @Override
    public List<Object> validValues(String s, Map<String, Object> map) {
      return (List) VALID_VALUES;
    }

    @Override
    public boolean visible(String s, Map<String, Object> map) {
      return true;
    }
  }

}
