// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import kafka.common.Topic;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.RoundRobinAssignor;
import org.apache.kafka.clients.consumer.internals.PartitionAssignor;
import org.apache.kafka.common.Cluster;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.errors.TimeoutException;
import org.apache.kafka.common.errors.WakeupException;
import org.apache.kafka.common.serialization.ByteArrayDeserializer;
import org.apache.kafka.connect.connector.ConnectorContext;
import org.apache.kafka.connect.errors.ConnectException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

/**
 * The purpose of this thread is to monitor the topics in the source cluster, detect
 * when there are changes affecting the topics to replicate, and assign partitions to
 * the tasks accordingly.
 */
public class TopicMonitorThread extends Thread {
  private static final Logger log = LoggerFactory.getLogger(ReplicatorSourceConnector.class);

  private final ConnectorContext context;
  private final Consumer<byte[], byte[]> consumer;
  private final Set<String> whitelistTopics;
  private final Pattern topicPattern;
  private final Set<String> blacklistTopics;
  private final String connectorName;
  private final long pollIntervalMs;

  private final PartitionAssignor assignor = new RoundRobinAssignor();
  private final CountDownLatch firstRefreshLatch = new CountDownLatch(1);
  private final CountDownLatch shutdownLatch = new CountDownLatch(1);

  private volatile Map<String, List<PartitionInfo>> currentTopics = null;

  public TopicMonitorThread(ConnectorContext context,
                            ReplicatorSourceConnectorConfig config) {
    this(context, config.getName(), config.getTopics(), config.getTopicPattern(),
            config.getBlacklistTopics(), config.getTopicPollIntervalMs(), buildConsumer(config));
  }

  TopicMonitorThread(ConnectorContext context,
                     String connectorName,
                     Set<String> whitelistTopics,
                     Pattern topicPattern,
                     Set<String> blacklistTopics,
                     int pollIntervalMs,
                     Consumer<byte[], byte[]> consumer) {
    super("topic-monitor-thread-" + connectorName);
    this.context = context;
    this.whitelistTopics = whitelistTopics;
    this.topicPattern = topicPattern;
    this.blacklistTopics = blacklistTopics;
    this.connectorName = connectorName;
    this.pollIntervalMs = pollIntervalMs;
    this.consumer = consumer;
  }

  private static KafkaConsumer<byte[], byte[]> buildConsumer(ReplicatorSourceConnectorConfig config) {
    Map<String, Object> consumerConfig = new HashMap<>();
    consumerConfig.putAll(config.getSourceConsumerConfigs());
    if (!consumerConfig.containsKey(ConsumerConfig.CLIENT_ID_CONFIG)) {
      consumerConfig.put(ConsumerConfig.CLIENT_ID_CONFIG, config.getName());
    }
    return new KafkaConsumer<>(consumerConfig, new ByteArrayDeserializer(), new ByteArrayDeserializer());
  }

  public Map<String, PartitionAssignor.Assignment> assignments(int maxTasks) {
    if (currentTopics == null) {
      try {
        firstRefreshLatch.await(10, TimeUnit.SECONDS);
      } catch (InterruptedException e) {
        // We shouldn't get any interrupts from Connect, but reset the interrupt flag just in case
        Thread.currentThread().interrupt();
      }
    }

    if (currentTopics == null)
      throw new ConnectException("Could not obtain timely topic metadata update from source cluster");

    Map<String, List<PartitionInfo>> currentTopics = this.currentTopics;
    if (currentTopics.isEmpty())
      return Collections.emptyMap();

    Cluster cluster = buildCluster(currentTopics);
    int numTasks = Math.min(numPartitionsToAssign(currentTopics), maxTasks);

    Map<String, PartitionAssignor.Subscription> subscriptions = buildSubscriptions(currentTopics, numTasks);
    Map<String, PartitionAssignor.Assignment> assignments = assignor.assign(cluster, subscriptions);

    log.debug("Finished computing task topic partition assignments: {}", assignments);

    return assignments;
  }

  private int numPartitionsToAssign(Map<String, List<PartitionInfo>> currentTopics) {
    int numPartitions = 0;
    for (List<PartitionInfo> partitions : currentTopics.values())
      numPartitions += partitions.size();
    return numPartitions;
  }

  private Map<String, PartitionAssignor.Subscription> buildSubscriptions(Map<String, List<PartitionInfo>> currentTopics, int numTasks) {
    List<String> topicsList = new ArrayList<>(currentTopics.keySet());
    Map<String, PartitionAssignor.Subscription> subscriptions = new HashMap<>(numTasks);

    for (int i = 0; i < numTasks; i++) {
      String taskId = connectorName + "-" + i;
      PartitionAssignor.Subscription subscription = new PartitionAssignor.Subscription(topicsList);
      subscriptions.put(taskId, subscription);
    }

    return subscriptions;
  }

  private Cluster buildCluster(Map<String, List<PartitionInfo>> currentTopics) {
    Set<Node> nodes = new HashSet<>();
    List<PartitionInfo> matchingPartitionInfo = new ArrayList<>();

    for (Map.Entry<String, List<PartitionInfo>> topicEntry : currentTopics.entrySet()) {
      matchingPartitionInfo.addAll(topicEntry.getValue());
      for (PartitionInfo partitionInfo : topicEntry.getValue())
        Collections.addAll(nodes, partitionInfo.replicas());
    }

    return new Cluster(
            null, // the cluster id doesn't matter for partition assignment
            nodes,
            matchingPartitionInfo,
            Collections.<String>emptySet(),
            Collections.<String>emptySet());
  }

  private boolean matchesTopicPattern(String topic) {
    return topicPattern != null && topicPattern.matcher(topic).matches();
  }

  private boolean isInternalTopic(String topic) {
    // TODO: It would be preferable to depend on the internal topic flag set in TopicMetadata response,
    // but this isn't exposed in the consumer's APIs. We should have it when we switch to AdminClient though.
    return Topic.InternalTopics().contains(topic);
  }

  private Map<String, List<PartitionInfo>> listMatchingTopics() {
    Map<String, List<PartitionInfo>> topicMetadata = consumer.listTopics();
    Map<String, List<PartitionInfo>> matchingTopics = new HashMap<>();

    for (Map.Entry<String, List<PartitionInfo>> topicEntry : topicMetadata.entrySet()) {
      String topic = topicEntry.getKey();

      if (blacklistTopics.contains(topic))
        continue;

      if (whitelistTopics.contains(topic) || (!isInternalTopic(topic) && matchesTopicPattern(topic)))
        matchingTopics.put(topic, topicEntry.getValue());
    }

    return matchingTopics;
  }

  private boolean matchesPartitions(Map<String, List<PartitionInfo>> topicMetadataA,
                                    Map<String, List<PartitionInfo>> topicMetadataB) {
    if (topicMetadataA.size() != topicMetadataB.size())
      return false;

    for (Map.Entry<String, List<PartitionInfo>> topicEntry : topicMetadataA.entrySet()) {
      List<PartitionInfo> oldMetadata = topicMetadataB.get(topicEntry.getKey());
      List<PartitionInfo> newMetadata = topicEntry.getValue();
      if (oldMetadata == null || newMetadata.size() != oldMetadata.size())
        return false;
    }

    return true;
  }

  public void shutdown() {
    // Wakeup the consumer in case we're blocked in listTopics()
    consumer.wakeup();
    shutdownLatch.countDown();
    synchronized (consumer) {
      consumer.close();
    }
  }

  public void run() {
    try {
      while (true) {
        try {
          synchronized (consumer) {
            if (shutdownLatch.getCount() == 0)
              return;

            Map<String, List<PartitionInfo>> updatedTopics = listMatchingTopics();
            Map<String, List<PartitionInfo>> currentTopics = this.currentTopics;

            this.currentTopics = updatedTopics;
            if (currentTopics == null)
              firstRefreshLatch.countDown();
            else if (!matchesPartitions(currentTopics, updatedTopics))
              context.requestTaskReconfiguration();
          }

          shutdownLatch.await(pollIntervalMs, TimeUnit.MILLISECONDS);
        } catch (TimeoutException e) {
          log.warn("Timeout refreshing topic metadata from the source cluster", e);
        } catch (WakeupException e) {
          log.debug("Topic monitor thread woken up for shutdown");
          return;
        }
      }
    } catch (Exception e) {
      log.error("Unexpected exception in topic monitor thread", e);
      context.raiseError(e);
    }
  }
}
