// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator.util;

import io.confluent.common.utils.Time;
import kafka.admin.AdminUtils;
import kafka.admin.BrokerMetadata;
import kafka.admin.RackAwareMode;
import kafka.server.ConfigType;
import kafka.utils.ZkUtils;
import org.I0Itec.zkclient.IZkChildListener;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.errors.TopicExistsException;
import org.apache.kafka.common.protocol.Errors;
import org.apache.kafka.common.requests.MetadataResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;
import scala.collection.Iterator;
import scala.collection.JavaConversions;
import scala.collection.Seq;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * This class is used to abstract some of the admin APIs to make testing easier. Eventually
 * we expect this to be replaced by the AdminClient that will be shipped with kafka-clients.
 *
 * Note on fetching topic metadata: we can either fetch topic metadata from Kafka or
 * Zookeeper. The risk of fetching from Kafka is that topics may be inadvertently created
 * if the user has enabled auto-creation of topics. Although we will certainly discourage
 * users from enabling this, we've chosen to fetch metadata from Zookeeper instead to avoid
 * the risk of misconfiguration.
 */
public class ReplicatorAdminClient implements Closeable {
  private static final Logger log = LoggerFactory.getLogger(ReplicatorAdminClient.class);
  private static final RackAwareMode RACK_AWARE_SAFE = RackAwareMode.Safe$.MODULE$;
  private static final String TOPIC_METADATA_PATH = ZkUtils.BrokerTopicsPath();
  private static final int FORCE_REFRESH_DEADLINE = 0;

  private final ZkUtils zkUtils;
  private final Time time;
  private final int maxAgeMs;

  private Collection<BrokerMetadata> brokers;
  private long nextBrokerMetadataRefresh;

  private Set<String> interestedTopics;
  private long nextTopicMetadataRefresh;
  private Map<String, TopicMetadata> topicMetadata;
  private TopicMetadataListener metadataListener;
  private IZkChildListener zkTopicMetadataListener;

  public ReplicatorAdminClient(ZkUtils zkUtils,
                               Time time,
                               int maxAgeMs) {
    this.zkUtils = zkUtils;
    this.time = time;
    this.maxAgeMs = maxAgeMs;
    this.topicMetadata = new HashMap<>();
    this.interestedTopics = Collections.emptySet();

    long now = time.milliseconds();
    this.nextTopicMetadataRefresh = now;
    this.nextBrokerMetadataRefresh = now;
  }

  public Properties topicConfig(String topic) {
    return AdminUtils.fetchEntityConfig(zkUtils, ConfigType.Topic(), topic);
  }

  /**
   * Change the configuration of the given topic.
   * @param topic the topic
   * @param config the new configuration
   */
  public void changeTopicConfig(String topic, Properties config) {
    AdminUtils.changeTopicConfig(zkUtils, topic, config);
  }

  /**
   * Get the number of currently active brokers in the cluster
   * @return the count of brokers
   */
  public synchronized int aliveBrokers() {
    long now = time.milliseconds();
    if (brokers != null && now < nextBrokerMetadataRefresh)
      return brokers.size();

    refreshBrokerMetadata();
    return brokers.size();
  }

  private void refreshBrokerMetadata() {
    Seq<BrokerMetadata> brokerMetadata = AdminUtils.getBrokerMetadatas(zkUtils,
            RACK_AWARE_SAFE, Option.<Seq<Object>>empty());
    brokers = JavaConversions.asJavaCollection(brokerMetadata);
    nextBrokerMetadataRefresh = time.milliseconds() + maxAgeMs;
  }

  /**
   * Set the list of topics to subscribe to metadata changes.
   * @param topics the set of topics to monitor metadata for
   * @param topicMetadataListener listener which will be notified whenever metadata for these topics is refreshed
   */
  public synchronized void setInterestedTopics(Set<String> topics, TopicMetadataListener topicMetadataListener) {
    if (zkTopicMetadataListener != null)
      zkUtils.zkClient().unsubscribeChildChanges(TOPIC_METADATA_PATH, zkTopicMetadataListener);

    this.metadataListener = topicMetadataListener;
    this.interestedTopics = topics;

    // request a metadata refresh if we don't have metadata for all topics
    topicMetadata.keySet().retainAll(topics);

    if (topicMetadata.size() < topics.size())
      nextTopicMetadataRefresh = FORCE_REFRESH_DEADLINE;

    zkTopicMetadataListener = new IZkChildListener() {
      @Override
      public void handleChildChange(String parentPath, List<String> topics) throws Exception {
        synchronized (ReplicatorAdminClient.this) {
          Set<String> topicsToRefresh = new HashSet<>();
          for (String topic : topics) {
            if (interestedTopics.contains(topic))
              topicsToRefresh.add(topic);
          }

          if (!topicsToRefresh.isEmpty())
            refreshTopicMetadata(topicsToRefresh);
        }
      }
    };

    zkUtils.zkClient().subscribeChildChanges(TOPIC_METADATA_PATH, zkTopicMetadataListener);
  }

  private Map<String, TopicMetadata> topicMetadata() {
    long now = time.milliseconds();
    if (now >= nextTopicMetadataRefresh)
      refreshTopicMetadata(interestedTopics);
    return topicMetadata;
  }

  /**
   * Get the current metadata available for a given topic.
   *
   * @param topic the topic to check metadata for
   * @return the current known metadata or null if none is available
   */
  public synchronized TopicMetadata topicMetadata(String topic) {
    return topicMetadata().get(topic);
  }

  /**
   * Check whether a partition exists.
   *
   * @param partition the partition to check
   * @return true if the partition is known to exist, false otherwise
   */
  public synchronized boolean partitionExists(TopicPartition partition) {
    TopicMetadata metadata = topicMetadata().get(partition.topic());
    if (metadata == null)
      return false;
    return partition.partition() < metadata.numPartitions();
  }

  /**
   * Check whether a topic exists.
   * @param topic
   * @return
   */
  public synchronized boolean topicExists(String topic) {
    return topicMetadata().get(topic) != null;
  }

  /**
   * Create a topic, blocking until the operation is completed successfully or fails. If the topic
   * exists (or is in the process of being created by another user), this method will return false.
   * If that happens, the caller should check
   *
   * @param topic name of the topic to create
   * @param numPartitions the number of partitions that the topic should have
   * @param replicationFactor expected replication factor
   * @param topicConfig initial topic configuration
   * @return true if the topic creation completed successfully, false if the topic already exists
   */
  public boolean createTopic(String topic,
                          int numPartitions,
                          int replicationFactor,
                          Properties topicConfig) {
    try {
      log.info("Creating topic {} with {} partitions, replication factor {}, and config {}",
              topic, numPartitions, replicationFactor, topicConfig);
      this.nextTopicMetadataRefresh = FORCE_REFRESH_DEADLINE;
      AdminUtils.createTopic(
              zkUtils,
              topic,
              numPartitions,
              replicationFactor,
              topicConfig,
              RACK_AWARE_SAFE);
      return true;
    } catch (TopicExistsException e) {
      log.debug("Failed to create topic {} because it already exists", topic, e);
      return false;
    }
  }

  /**
   * Increase the number of partitions.
   *
   * @param topic the topic to add partitions to
   * @param numPartitions the expected number of partitions
   */
  public void addPartitions(String topic,
                            int numPartitions) {
    AdminUtils.addPartitions(
            zkUtils,
            topic,
            numPartitions,
            null, // let kafka determine the replica assignment
            true, // ensure assigned replica is available
            RACK_AWARE_SAFE);
    this.nextTopicMetadataRefresh = FORCE_REFRESH_DEADLINE;
  }

  private void refreshTopicMetadata(Set<String> topics) {
    scala.collection.Set<MetadataResponse.TopicMetadata> allTopicMetadata = AdminUtils.fetchTopicMetadataFromZk(
            JavaConversions.asScalaSet(topics), zkUtils);
    Iterator<MetadataResponse.TopicMetadata> iterator = allTopicMetadata.iterator();

    topicMetadata.clear();
    while (iterator.hasNext()) {
      MetadataResponse.TopicMetadata topicMetadataResponse = iterator.next();
      if (topicMetadataResponse.error() == Errors.NONE) {
        TopicMetadata nextTopicMetadata = convertTopicMetadata(topicMetadataResponse);
        topicMetadata.put(nextTopicMetadata.topic(), nextTopicMetadata);
      }
    }

    nextTopicMetadataRefresh = time.milliseconds() + maxAgeMs;

    if (metadataListener != null)
      metadataListener.onTopicMetadataRefresh();
  }

  private static TopicMetadata convertTopicMetadata(MetadataResponse.TopicMetadata topicMetadata) {
    String topic = topicMetadata.topic();
    List<PartitionInfo> partitions = new ArrayList<>();
    for (MetadataResponse.PartitionMetadata partitionMetadata : topicMetadata.partitionMetadata()) {
      int partition = partitionMetadata.partition();
      Node leader = partitionMetadata.leader();
      Node[] isr = partitionMetadata.isr().toArray(new Node[0]);
      Node[] replicas = partitionMetadata.replicas().toArray(new Node[0]);
      partitions.add(new PartitionInfo(topic, partition, leader, replicas, isr));
    }
    return new TopicMetadata(topic, partitions);
  }

  public void close() {
    zkUtils.close();
  }

  /**
   * Used to get notifications when metadata has been refreshed
   */
  public interface TopicMetadataListener {

    /**
     * Invoked on every topic metadata refresh.
     */
    void onTopicMetadataRefresh();

  }

}
