// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import io.confluent.common.utils.MockTime;
import io.confluent.connect.replicator.util.ByteArrayConverter;
import io.confluent.connect.replicator.util.ReplicatorAdminClient;
import io.confluent.connect.replicator.util.ReplicatorAdminClient.TopicMetadataListener;
import io.confluent.connect.replicator.util.TopicMetadata;
import io.confluent.connect.replicator.util.Utils;
import kafka.log.LogConfig;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.OffsetOutOfRangeException;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.record.TimestampType;
import org.apache.kafka.connect.source.SourceRecord;
import org.apache.kafka.connect.source.SourceTaskContext;
import org.apache.kafka.connect.storage.Converter;
import org.apache.kafka.connect.storage.OffsetStorageReader;
import org.easymock.Capture;
import org.easymock.EasyMock;
import org.easymock.EasyMockSupport;
import org.junit.Before;
import org.junit.Test;
import org.junit.experimental.theories.DataPoint;
import org.junit.experimental.theories.Theories;
import org.junit.experimental.theories.Theory;
import org.junit.runner.RunWith;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import static io.confluent.connect.replicator.util.Utils.toConnectOffset;
import static io.confluent.connect.replicator.util.Utils.toConnectPartition;
import static java.util.Collections.emptySet;
import static java.util.Collections.singleton;
import static java.util.Collections.singletonList;
import static java.util.Collections.singletonMap;
import static org.easymock.EasyMock.anyLong;
import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.expectLastCall;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

@RunWith(Theories.class)
public class ReplicatorSourceTaskTest extends EasyMockSupport {

  private final String taskId = "replicator-1";
  private final Converter byteArrayConverter = new ByteArrayConverter();

  private MockTime time;

  private static final TimestampType DEFAULT_TIMESTAMP_TYPE = TimestampType.CREATE_TIME;

  @DataPoint
  public static TimestampType[] typeParams() {
    return new TimestampType[] {
        TimestampType.NO_TIMESTAMP_TYPE,
        TimestampType.CREATE_TIME,
        TimestampType.LOG_APPEND_TIME
    };
  }

  @DataPoint
  public static TimestampType[] userDefinedTypeParams() {
    return new TimestampType[] {
        TimestampType.CREATE_TIME,
        TimestampType.LOG_APPEND_TIME,
        null
    };
  }
  @Before
  public void setup() {
    this.time = new MockTime();
  }

  @Test
  public void testVersion() {
    SourceTaskContext context = createMock(SourceTaskContext.class);
    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(false);

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    String version = task.version();
    assertNotNull(version);
    assertFalse(version.isEmpty());

    verifyAll();
  }

  @Test
  public void testConsumerRecordWithNoTimestamp() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);

    ConsumerRecords<byte[], byte[]> records = new ConsumerRecords<>(singletonMap(sourcePartition,
            Arrays.asList(new ConsumerRecord<>(sourceTopic, 0, 0L, ConsumerRecord.NO_TIMESTAMP, TimestampType.NO_TIMESTAMP_TYPE,
                    0, 0, 0, new byte[0], new byte[0]))));

    List<SourceRecord> sourceRecords =
        expectSimpleReceive(sourceTopic, sourcePartition, topicRenameFormat, records, null);
    assertEquals(1, sourceRecords.size());

    SourceRecord sourceRecord = sourceRecords.get(0);
    assertNull(sourceRecord.timestamp());

    verifyAll();
  }

  @Test
  public void testPartitionPreservationDisabled() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);

    ConsumerRecords<byte[], byte[]> records = new ConsumerRecords<>(singletonMap(sourcePartition,
            Arrays.asList(new ConsumerRecord<>(sourceTopic, 0, 0L, ConsumerRecord.NO_TIMESTAMP, TimestampType.NO_TIMESTAMP_TYPE,
                    0, 0, 0, new byte[0], new byte[0]))));

    List<SourceRecord> sourceRecords =
            expectSimpleReceive(sourceTopic, sourcePartition, topicRenameFormat, records, null, false);
    assertEquals(1, sourceRecords.size());

    SourceRecord sourceRecord = sourceRecords.get(0);
    assertNull(sourceRecord.kafkaPartition());

    verifyAll();
  }

  @Theory
  public void testTimestampTypeOverride(TimestampType timestampType, TimestampType userDefinedType) throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);

    long timestamp = 500L;
    ConsumerRecords<byte[], byte[]> records = new ConsumerRecords<>(singletonMap(sourcePartition,
        Arrays.asList(new ConsumerRecord<>(sourceTopic, 0, 0L, timestamp, timestampType,
            0, 0, 0, new byte[0], new byte[0]))));

    expectSimpleReceive(sourceTopic, sourcePartition, topicRenameFormat, records, userDefinedType);
    verifyAll();
  }

  @Test
  public void testTopicRename() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);

    long timestamp = 500L;
    ConsumerRecords<byte[], byte[]> records = new ConsumerRecords<>(singletonMap(sourcePartition,
            Arrays.asList(new ConsumerRecord<>(sourceTopic, 0, 0L, timestamp, TimestampType.CREATE_TIME,
                    0, 0, 0, new byte[0], new byte[0]))));

    List<SourceRecord> sourceRecords =
        expectSimpleReceive(sourceTopic, sourcePartition, topicRenameFormat, records, null);
    assertEquals(1, sourceRecords.size());

    SourceRecord sourceRecord = sourceRecords.get(0);
    assertEquals("dc.foo", sourceRecord.topic());
    assertEquals(0, sourceRecord.kafkaPartition().intValue());
    assertEquals(timestamp, sourceRecord.timestamp().longValue());

    Map<String, ?> connectPartition = sourceRecord.sourcePartition();
    assertEquals(sourceTopic, connectPartition.get("topic"));
    assertEquals(0, connectPartition.get("partition"));

    verifyAll();
  }

  @Test
  public void testNonExistingTopicCreation() throws Exception {
    // if the task is assigned partition 0, it should create the the topic

    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";
    int topicConfigSyncIntervalMs = 10000;

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);
    TopicPartition destPartition = new TopicPartition(destTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicCreateBackoffMs()).andStubReturn(10000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    Properties topicConfig = new Properties();
    expect(sourceClient.topicConfig(sourceTopic)).andReturn(topicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(destClient.topicMetadata(destTopic)).andReturn(null);
    expect(destClient.partitionExists(destPartition)).andReturn(false);

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    consumer.pause(singleton(sourcePartition));
    expectLastCall();

    // create the topic
    expect(destClient.topicMetadata(destTopic)).andReturn(null);
    expect(destClient.aliveBrokers()).andReturn(1);
    expect(destClient.createTopic(destTopic, 1, 1, topicConfig)).andReturn(true);
    expect(config.getTopicCreateBackoffMs()).andReturn(10000);

    // now the topic should exist
    expect(destClient.topicExists(destTopic)).andReturn(true);

    expect(sourceClient.topicConfig(sourceTopic)).andReturn(new Properties());
    Properties sourceTopicConfig = new Properties();
    overrideTimestampType(DEFAULT_TIMESTAMP_TYPE, sourceTopicConfig);
    destClient.changeTopicConfig(destTopic, sourceTopicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expectLastCall();
    expect(destClient.topicConfig(destTopic)).andReturn(new Properties());
    expect(config.getTopicConfigSyncIntervalMs()).andReturn(topicConfigSyncIntervalMs);

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    final ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testUpdateTopicConfiguration() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);
    TopicPartition destPartition = new TopicPartition(destTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);

    consumer.assign(sourceAssignment);
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    Properties sourceTopicConfig = new Properties();
    sourceTopicConfig.put("max.message.bytes", "2000000");

    expect(offsetReader.offset(toConnectPartition(sourceTopic, 0))).andReturn(null);
    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    expect(destClient.topicMetadata(destTopic)).andStubReturn(new TopicMetadata(
            destTopic, singletonList(new PartitionInfo(destTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }))));
    expect(destClient.aliveBrokers()).andStubReturn(1);
    expect(destClient.partitionExists(destPartition)).andReturn(true);
    expect(destClient.topicExists(destTopic)).andReturn(true);

    expect(sourceClient.topicConfig(sourceTopic)).andReturn(sourceTopicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(destClient.topicConfig(destTopic)).andReturn(new Properties());
    destClient.changeTopicConfig(destTopic, sourceTopicConfig);
    expectLastCall();

    replayAll();

    final ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testConfigNotQueriedWhenAutoSyncDisabled() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";
    int partitionId = 0;

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, partitionId);
    TopicPartition destPartition = new TopicPartition(destTopic, partitionId);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(false);

    consumer.assign(sourceAssignment);
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, partitionId, leader,
                                                                                              new Node[]{ leader }, new Node[]{ leader })));

    expect(offsetReader.offset(toConnectPartition(sourceTopic, partitionId))).andReturn(null);
    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    expect(destClient.topicMetadata(destTopic)).andStubReturn(new TopicMetadata(
        destTopic, singletonList(new PartitionInfo(destTopic, partitionId, leader, new Node[]{leader }, new Node[]{leader }))));
    expect(destClient.partitionExists(destPartition)).andReturn(true);

    replayAll();

    final ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testSkipTopicCreationIfNotEnoughBrokers() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);
    TopicPartition destPartition = new TopicPartition(destTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicCreateBackoffMs()).andReturn(10000);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);

    consumer.assign(sourceAssignment);
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    Node replica = new Node(0, "localhost", 9093);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader, replica }, new Node[]{ leader, replica })));

    Properties topicConfig = new Properties();
    expect(sourceClient.topicConfig(sourceTopic)).andReturn(topicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(destClient.topicMetadata(destTopic)).andStubReturn(null);
    expect(destClient.aliveBrokers()).andReturn(1);

    expect(destClient.partitionExists(destPartition)).andReturn(false);
    expect(destClient.topicExists(destTopic)).andReturn(false);

    consumer.pause(singleton(sourcePartition));
    expectLastCall();

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    replayAll();

    final ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testTopicCreationRetry() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";
    int topicCreateBackoffMs = 5000;
    int topicConfigSyncIntervalMs = 10000;

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);
    TopicPartition destPartition = new TopicPartition(destTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(config.getTopicPreservePartitions()).andStubReturn(true);

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    consumer.pause(singleton(sourcePartition));
    expectLastCall();

    // try to create the topic
    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    Properties topicConfig = new Properties();
    expect(destClient.createTopic(destTopic, 1, 1, topicConfig)).andThrow(new RuntimeException());
    expect(config.getTopicCreateBackoffMs()).andStubReturn(topicCreateBackoffMs);

    expect(destClient.topicExists(destTopic)).andStubReturn(false);
    expect(sourceClient.topicConfig(sourceTopic)).andStubReturn(topicConfig);
    expect(destClient.topicMetadata(destTopic)).andStubReturn(null);
    expect(destClient.aliveBrokers()).andStubReturn(1);
    expect(destClient.partitionExists(destPartition)).andStubReturn(false);
    expect(config.getTopicConfigSyncIntervalMs()).andReturn(topicConfigSyncIntervalMs);

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    final ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();

    resetAll();

    // verify we don't retry if we're still in the backoff window

    expect(config.getTopicCreateBackoffMs()).andStubReturn(topicCreateBackoffMs);
    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(config.getTopicPreservePartitions()).andStubReturn(true);

    expect(destClient.partitionExists(destPartition)).andStubReturn(false);
    expect(destClient.topicExists(destTopic)).andStubReturn(false);

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    time.sleep(topicCreateBackoffMs/2);
    assertTrue(task.poll().isEmpty());

    verifyAll();

    resetAll();

    expect(destClient.partitionExists(destPartition)).andStubReturn(false);

    time.sleep(topicCreateBackoffMs);
    // verify we do retry topic creation after the backoff window completes
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);
    expect(config.getTopicCreateBackoffMs()).andStubReturn(0);
    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicAutoCreate()).andStubReturn(true);

    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    expect(sourceClient.topicConfig(sourceTopic)).andStubReturn(topicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expect(destClient.aliveBrokers()).andStubReturn(1);

    expect(destClient.createTopic(destTopic, 1, 1, topicConfig)).andReturn(true);
    expect(destClient.topicMetadata(destTopic)).andReturn(null);

    expect(consumer.paused()).andReturn(singleton(sourcePartition));
    expect(destClient.partitionExists(destPartition)).andReturn(true);

    consumer.resume(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.paused()).andReturn(emptySet());

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testSeekCommittedOffset() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 1);
    TopicPartition destPartition = new TopicPartition(destTopic, 1);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(consumer.paused()).andStubReturn(emptySet());

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    // the partition exists and we seek to the committed offset (+ 1)
    expect(destClient.partitionExists(destPartition)).andStubReturn(true);
    expect(offsetReader.offset(toConnectPartition(sourcePartition))).andReturn(toConnectOffset(50L));
    consumer.seek(sourcePartition, 51L);
    expectLastCall();

    // poll causes an out of range error and a seek to the beginning
    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testOffsetOutOfRangeHandling() throws Exception {
    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 1);
    TopicPartition destPartition = new TopicPartition(destTopic, 1);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(consumer.paused()).andStubReturn(emptySet());

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    expect(destClient.partitionExists(destPartition)).andStubReturn(true);
    expect(offsetReader.offset(toConnectPartition(sourcePartition))).andReturn(null);

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    // poll causes an out of range error and a seek to the beginning
    expect(consumer.poll(anyLong())).andThrow(new OffsetOutOfRangeException(singletonMap(sourcePartition, 51L)));

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testExpandingTopicPartitions() throws Exception {
    // if the task is assigned partition 0, it should expand partitions whenever needed

    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);
    TopicPartition destPartition = new TopicPartition(destTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);
    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicCreateBackoffMs()).andStubReturn(10000);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);

    consumer.assign(sourceAssignment);
    expectLastCall();

    expect(consumer.paused()).andStubReturn(emptySet());
    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(Arrays.asList(
            new PartitionInfo(sourceTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(sourceTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(destClient.topicMetadata(destTopic)).andStubReturn(new TopicMetadata(
            destTopic, singletonList(new PartitionInfo(destTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }))));

    destClient.addPartitions(destTopic, 2);
    expectLastCall();

    expect(destClient.partitionExists(destPartition)).andReturn(true);
    expect(destClient.topicExists(destTopic)).andReturn(true);
    expect(offsetReader.offset(toConnectPartition(sourcePartition))).andReturn(null);

    Properties sourceTopicConfig = new Properties();
    expect(sourceClient.topicConfig(sourceTopic)).andReturn(new Properties());
    overrideTimestampType(DEFAULT_TIMESTAMP_TYPE, sourceTopicConfig);
    destClient.changeTopicConfig(destTopic, sourceTopicConfig);
    expect(config.getTopicTimestampType()).andStubReturn(DEFAULT_TIMESTAMP_TYPE.toString());
    expectLastCall();
    expect(destClient.topicConfig(destTopic)).andReturn(new Properties());

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testSkipPartitionExpansionIfNotPartitionsPreserved() throws Exception {
    // if the task is assigned partition 0, it should expand partitions whenever needed

    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 0);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(false); // disable partition preservation
    expect(config.getTopicCreateBackoffMs()).andStubReturn(10000);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(false);

    consumer.assign(sourceAssignment);
    expectLastCall();

    expect(consumer.paused()).andStubReturn(emptySet());
    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(Arrays.asList(
            new PartitionInfo(sourceTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(sourceTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(destClient.topicMetadata(destTopic)).andStubReturn(new TopicMetadata(
            destTopic, singletonList(new PartitionInfo(destTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }))));
    expect(destClient.topicExists(destTopic)).andStubReturn(true);

    expect(offsetReader.offset(toConnectPartition(sourcePartition))).andReturn(null);
    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testNonExistingTopicPaused() throws Exception {
    // if the task is not assigned partition 0, it just pauses the partition and awaits creation

    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, 1);
    TopicPartition destPartition = new TopicPartition(destTopic, 1);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(true);

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    Capture<TopicMetadataListener> capturedListener = new Capture<>();
    destClient.setInterestedTopics(eq(singleton(destTopic)), EasyMock.capture(capturedListener));
    expectLastCall();

    // the topic doesn't exist and we shouldn't try to create it
    expect(destClient.partitionExists(destPartition)).andStubReturn(false);

    consumer.pause(singleton(sourcePartition));
    expectLastCall();

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
    resetAll();

    time.sleep(ReplicatorSourceTask.RETRY_RESUME_FREQUENCY_MS);

    // the topic is created externally and our metadata cache updates
    expect(destClient.partitionExists(destPartition)).andReturn(true);

    expect(context.offsetStorageReader()).andStubReturn(offsetReader);
    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(consumer.paused()).andReturn(singleton(sourcePartition));

    consumer.resume(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.paused()).andReturn(emptySet());

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  @Test
  public void testNonExistentPartitionsPausedWhenAutoCreateDisabled() throws Exception {
    // if the task is assigned partition 0 but autoTopicCreate -> false, it just pauses the partition and awaits creation

    String sourceTopic = "foo";
    String topicRenameFormat = "dc.${topic}";
    String destTopic = "dc.foo";
    int partitionId = 0;

    TopicPartition sourcePartition = new TopicPartition(sourceTopic, partitionId);
    TopicPartition destPartition = new TopicPartition(destTopic, partitionId);

    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(true);
    expect(config.getTopicAutoCreate()).andStubReturn(false);
    expect(config.getTopicConfigSync()).andStubReturn(false);

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andStubReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    expect(destClient.topicMetadata(destTopic)).andStubReturn(null);
    expect(destClient.partitionExists(destPartition)).andStubReturn(false);

    // initial assignment and topic metadata checks
    consumer.assign(sourceAssignment);
    expectLastCall();

    Capture<TopicMetadataListener> capturedListener = new Capture<>();
    destClient.setInterestedTopics(eq(singleton(destTopic)), EasyMock.capture(capturedListener));
    expectLastCall();

    // the topic doesn't exist and we shouldn't try to create it
    expect(destClient.partitionExists(destPartition)).andStubReturn(false);
    expect(destClient.topicExists(destTopic)).andStubReturn(false);

    consumer.pause(singleton(sourcePartition));
    expectLastCall();

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);
    assertTrue(task.poll().isEmpty());

    verifyAll();
    resetAll();

    time.sleep(ReplicatorSourceTask.RETRY_RESUME_FREQUENCY_MS);

    // the topic is created externally and our metadata cache updates
    expect(destClient.partitionExists(destPartition)).andReturn(true);

    expect(context.offsetStorageReader()).andStubReturn(offsetReader);
    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(consumer.paused()).andReturn(singleton(sourcePartition));

    consumer.resume(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.paused()).andReturn(emptySet());

    expect(consumer.poll(anyLong())).andReturn(ConsumerRecords.empty());

    replayAll();

    assertTrue(task.poll().isEmpty());

    verifyAll();
  }

  private List<SourceRecord> expectSimpleReceive(String sourceTopic,
                                                 TopicPartition sourcePartition,
                                                 String topicRenameFormat,
                                                 ConsumerRecords<byte[], byte[]> records,
                                                 TimestampType userDefinedType) throws InterruptedException {
    return expectSimpleReceive(sourceTopic, sourcePartition, topicRenameFormat, records, userDefinedType, true);
  }

  private List<SourceRecord> expectSimpleReceive(String sourceTopic,
                                                 TopicPartition sourcePartition,
                                                 String topicRenameFormat,
                                                 ConsumerRecords<byte[], byte[]> records,
                                                 TimestampType userDefinedType,
                                                 boolean preservePartitions) throws InterruptedException {
    String destTopic = Utils.renameTopic(topicRenameFormat, sourceTopic);

    TopicPartition destPartition = new TopicPartition(destTopic, 0);
    List<TopicPartition> sourceAssignment = singletonList(sourcePartition);

    SourceTaskContext context = createMock(SourceTaskContext.class);
    OffsetStorageReader offsetReader = createMock(OffsetStorageReader.class);
    expect(context.offsetStorageReader()).andStubReturn(offsetReader);

    ReplicatorSourceTaskConfig config = createMock(ReplicatorSourceTaskConfig.class);
    Consumer consumer = createMock(Consumer.class);
    ReplicatorAdminClient sourceClient = createMock(ReplicatorAdminClient.class);
    ReplicatorAdminClient destClient = createMock(ReplicatorAdminClient.class);

    expect(config.getTopicRenameFormat()).andStubReturn(topicRenameFormat);
    expect(config.getTopicPreservePartitions()).andStubReturn(preservePartitions);
    expect(config.getTopicConfigSyncIntervalMs()).andStubReturn(120000);
    expect(config.getTopicAutoCreate()).andStubReturn(true);
    expect(config.getTopicConfigSync()).andStubReturn(true);

    destClient.setInterestedTopics(eq(singleton(destTopic)), anyObject(TopicMetadataListener.class));
    expectLastCall();

    consumer.assign(sourceAssignment);
    expectLastCall();

    Node leader = new Node(0, "localhost", 9092);
    expect(consumer.partitionsFor(sourceTopic)).andReturn(singletonList(new PartitionInfo(sourceTopic, 0, leader,
            new Node[]{ leader }, new Node[]{ leader })));

    expect(destClient.topicMetadata(destTopic)).andReturn(new TopicMetadata(
            destTopic, singletonList(new PartitionInfo(destTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }))));

    if (preservePartitions)
      expect(destClient.partitionExists(destPartition)).andReturn(true);
    expect(destClient.topicExists(destTopic)).andReturn(true);

    expect(sourceClient.topicConfig(sourceTopic)).andReturn(new Properties());
    Properties sourceTopicConfig = new Properties();
    TimestampType type = userDefinedType != null ? userDefinedType : DEFAULT_TIMESTAMP_TYPE;
    overrideTimestampType(type, sourceTopicConfig);
    destClient.changeTopicConfig(destTopic, sourceTopicConfig);
    expect(config.getTopicTimestampType()).andReturn(type.toString());
    expectLastCall();
    expect(destClient.topicConfig(destTopic)).andReturn(new Properties());

    expect(offsetReader.offset(toConnectPartition(sourceTopic, 0))).andReturn(null);

    consumer.seekToBeginning(singleton(sourcePartition));
    expectLastCall();

    expect(consumer.poll(anyLong())).andReturn(records);

    replayAll();

    ReplicatorSourceTask task = new ReplicatorSourceTask(config, context, taskId, time,
            consumer, sourceClient, destClient, byteArrayConverter, byteArrayConverter);

    task.doStart(sourceAssignment);

    return task.poll();
  }

  private void overrideTimestampType(TimestampType type, Properties config) {
    config.setProperty(LogConfig.MessageTimestampTypeProp(), type.toString());
  }

}
