// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator;

import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.internals.PartitionAssignor;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.connect.connector.ConnectorContext;
import org.easymock.EasyMockSupport;
import org.junit.Test;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import static java.util.Collections.singleton;
import static java.util.Collections.singletonList;
import static java.util.Collections.singletonMap;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.expectLastCall;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

public class TopicMonitorThreadTest extends EasyMockSupport {

  @Test
  public void testBasicAssignment() {
    String topic = "topic";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);
    Set<String> topics = singleton(topic);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", topics,
            null, Collections.<String>emptySet(), 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    expect(consumer.listTopics()).andReturn(singletonMap(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader }))));

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(topic, 0)));
    assertTrue(partitions.contains(new TopicPartition(topic, 1)));

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testEmptyAssignment() {
    String topic = "topic";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);
    Set<String> topics = singleton(topic);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", topics,
            null, Collections.<String>emptySet(), 10000, consumer);

    expect(consumer.listTopics()).andReturn(Collections.emptyMap());

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(0, assignments.size());

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testNoEmptyTaskAssignments() {
    String topic = "topic";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);
    Set<String> topics = singleton(topic);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", topics,
            null, Collections.<String>emptySet(), 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    expect(consumer.listTopics()).andReturn(singletonMap(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader }))));

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    // max tasks set higher than the number of partitions
    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(5);
    assertEquals(2, assignments.size());

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testRegexAssignment() {
    String matchingTopic = "foo";
    String nonMatchingTopic = "boo";

    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", Collections.<String>emptySet(),
            Pattern.compile("f.*"), Collections.<String>emptySet(), 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    Map<String, List<PartitionInfo>> topicMetadata = new HashMap<>();
    topicMetadata.put(matchingTopic, Arrays.asList(
            new PartitionInfo(matchingTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(matchingTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));
    topicMetadata.put(nonMatchingTopic, singletonList(
            new PartitionInfo(nonMatchingTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(consumer.listTopics()).andReturn(topicMetadata);

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(matchingTopic, 0)));
    assertTrue(partitions.contains(new TopicPartition(matchingTopic, 1)));

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testWhitelistInternalTopic() {
    String topic = "__consumer_offsets";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);
    Set<String> topics = singleton(topic);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", topics,
            null, Collections.<String>emptySet(), 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    expect(consumer.listTopics()).andReturn(singletonMap(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader }))));

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(topic, 0)));
    assertTrue(partitions.contains(new TopicPartition(topic, 1)));

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testExcludeInternalTopics() {
    String matchingTopic = "foo";
    String internalTopic = "__consumer_offsets";

    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", Collections.<String>emptySet(),
            Pattern.compile(".*"), Collections.<String>emptySet(), 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    Map<String, List<PartitionInfo>> topicMetadata = new HashMap<>();
    topicMetadata.put(matchingTopic, Arrays.asList(
            new PartitionInfo(matchingTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(matchingTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));
    topicMetadata.put(internalTopic, singletonList(
            new PartitionInfo(internalTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(consumer.listTopics()).andReturn(topicMetadata);

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(matchingTopic, 0)));
    assertTrue(partitions.contains(new TopicPartition(matchingTopic, 1)));

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testBlacklistTopicsFiltered() {
    String topic = "topic";
    String blackListTopic = "blacklisted";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);
    Set<String> blackListTopics = singleton(blackListTopic);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", Collections.<String>emptySet(),
            Pattern.compile(".*"), blackListTopics, 10000, consumer);

    Node leader = new Node(0, "localhost", 9092);

    Map<String, List<PartitionInfo>> topicMetadata = new HashMap<>();
    topicMetadata.put(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));
    topicMetadata.put(blackListTopic, Arrays.asList(
            new PartitionInfo(blackListTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(blackListTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(consumer.listTopics()).andReturn(topicMetadata);

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(topic, 0)));
    assertTrue(partitions.contains(new TopicPartition(topic, 1)));

    monitorThread.shutdown();

    verifyAll();
  }

  @Test
  public void testMetadataChangeTriggersReconfiguration() throws Exception {
    String topic = "topic";
    String otherTopic = "other";
    ConnectorContext context = createMock(ConnectorContext.class);
    Consumer consumer = createMock(Consumer.class);

    TopicMonitorThread monitorThread = new TopicMonitorThread(context, "connector", Collections.<String>emptySet(),
            Pattern.compile(".*"), Collections.<String>emptySet(), 100, consumer);

    Node leader = new Node(0, "localhost", 9092);

    Map<String, List<PartitionInfo>> topicMetadata = new HashMap<>();
    topicMetadata.put(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));

    Map<String, List<PartitionInfo>> refreshedTopicMetadata = new HashMap<>();
    refreshedTopicMetadata.put(topic, Arrays.asList(
            new PartitionInfo(topic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(topic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));
    refreshedTopicMetadata.put(otherTopic, Arrays.asList(
            new PartitionInfo(otherTopic, 0, leader, new Node[]{ leader }, new Node[]{ leader }),
            new PartitionInfo(otherTopic, 1, leader, new Node[]{ leader }, new Node[]{ leader })));

    expect(consumer.listTopics())
            .andReturn(topicMetadata)
            .andStubReturn(refreshedTopicMetadata);

    context.requestTaskReconfiguration();
    expectLastCall();

    consumer.wakeup();
    expectLastCall();

    consumer.close();
    expectLastCall();

    replayAll();

    monitorThread.start();

    Map<String, PartitionAssignor.Assignment> assignments = monitorThread.assignments(2);
    assertEquals(2, assignments.size());

    Set<TopicPartition> partitions = verifyAndCollectAssignedPartitions(assignments.values());
    assertEquals(2, partitions.size());
    assertTrue(partitions.contains(new TopicPartition(topic, 0)));
    assertTrue(partitions.contains(new TopicPartition(topic, 1)));

    Thread.sleep(200);
    monitorThread.shutdown();

    verifyAll();
  }

  private Set<TopicPartition> verifyAndCollectAssignedPartitions(Collection<PartitionAssignor.Assignment> assignments) {
    Set<TopicPartition> partitions = new HashSet<>();
    for (PartitionAssignor.Assignment assignment : assignments) {
      for (TopicPartition partition : assignment.partitions()) {
        assertFalse(partitions.contains(partition));
        partitions.add(partition);
      }
    }
    return partitions;
  }

}
