// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.connect.replicator.util;

import com.google.common.base.Predicate;
import com.google.common.io.Files;
import io.confluent.common.utils.SystemTime;
import io.confluent.connect.replicator.util.ReplicatorAdminClient.TopicMetadataListener;
import kafka.server.KafkaConfig;
import kafka.server.KafkaServerStartable;
import kafka.utils.ZkUtils;
import org.apache.curator.test.TestingServer;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;

import static java.util.Collections.singleton;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

public class ReplicatorAdminClientTest {

  private File kafkaLogDir;
  private TestingServer zkServer;
  private KafkaServerStartable kafkaServer;
  private ReplicatorAdminClient client;

  @Before
  public void startup() throws Exception {
    zkServer = new TestingServer();

    kafkaLogDir = Files.createTempDir();
    kafkaLogDir.deleteOnExit();

    Properties config = new Properties();
    config.put(KafkaConfig.ZkConnectProp(), zkServer.getConnectString());
    config.put(KafkaConfig.ListenersProp(), "PLAINTEXT://localhost:0");
    config.put(KafkaConfig.LogDirProp(), kafkaLogDir.getAbsolutePath());

    kafkaServer = new KafkaServerStartable(KafkaConfig.fromProps(config));
    kafkaServer.startup();

    ZkUtils zkUtils = ZkUtils.apply(zkServer.getConnectString(), 6000, 10000, false);
    client = new ReplicatorAdminClient(zkUtils, new SystemTime(), 60000);
  }

  @After
  public void shutdown() throws Exception {
    if (client != null)
      client.close();

    if (kafkaServer != null) {
      kafkaServer.shutdown();
      kafkaServer.awaitShutdown();
    }

    if (zkServer != null)
      zkServer.close();
  }

  @Test
  public void testQueryBrokerMetadata() {
    ZkUtils zkUtils = ZkUtils.apply(zkServer.getConnectString(), 6000, 10000, false);
    ReplicatorAdminClient client = new ReplicatorAdminClient(zkUtils, new SystemTime(), 60000);
    assertEquals(1, client.aliveBrokers());
  }

  @Test
  public void testQueryNonExistingTopic() {
    ZkUtils zkUtils = ZkUtils.apply(zkServer.getConnectString(), 6000, 10000, false);
    ReplicatorAdminClient client = new ReplicatorAdminClient(zkUtils, new SystemTime(), 60000);
    client.setInterestedTopics(singleton("foo"), new TopicMetadataListener() {
      @Override
      public void onTopicMetadataRefresh() {
      }
    });
    assertNull(client.topicMetadata("foo"));
  }

  @Test
  public void testCreateTopic() throws Exception {
    int numPartitions = 4;
    assertTrue(client.createTopic("foo", numPartitions, 1, new Properties()));

    client.setInterestedTopics(singleton("foo"), new TopicMetadataListener() {
      @Override
      public void onTopicMetadataRefresh() {
      }
    });

    TopicMetadata topicMetadata = client.topicMetadata("foo");
    assertNotNull(topicMetadata);
    assertEquals(numPartitions, topicMetadata.numPartitions());
  }

  @Test
  public void testCreatedTopicNotification() throws Exception {
    final AtomicBoolean notified = new AtomicBoolean(false);
    client.setInterestedTopics(singleton("foo"), new TopicMetadataListener() {
      @Override
      public void onTopicMetadataRefresh() {
        notified.set(true);
      }
    });

    assertTrue(client.createTopic("foo", 1, 1, new Properties()));

    waitUntilTrue("Failed to get notification of topic creation", 5000L, new Predicate<Void>() {
      @Override
      public boolean apply(Void aVoid) {
        return notified.get();
      }
    });
  }

  @Test
  public void testNoNotificationForUninterestedTopicCreation() throws Exception {
    final AtomicBoolean notified = new AtomicBoolean(false);
    client.setInterestedTopics(singleton("foo"), new TopicMetadataListener() {
      @Override
      public void onTopicMetadataRefresh() {
        notified.set(true);
      }
    });

    assertTrue(client.createTopic("bar", 1, 1, new Properties()));

    Thread.sleep(1000);
    assertFalse(notified.get());
  }

  private static void waitUntilTrue(String message, long timeout, Predicate<Void> predicate) throws InterruptedException {
    long deadline = System.currentTimeMillis() + timeout;
    while (System.currentTimeMillis() < deadline) {
      if (predicate.apply(null))
        return;
      Thread.sleep(50);
    }
    fail(message);
  }

}
